from .interfaces.stream import (
    DXF_stream_to_unstruc,
    OMF_stream_to_unstruc,
    CSV_wells_stream_to_unstruc,
    CSV_mesh_stream_to_unstruc,
    CSV_volume_stream_to_unstruc,
    CSV_volume_stream_to_struct,
    VTK_stream_to_struct,  
    MX_stream_to_unstruc,
    OBJ_stream_to_trisurf,
    GLTF_stream_to_trisurf,
    MSH_stream_to_struct,
)
